;; Copyright (c) 2006-2007 Carlos Ungil

;; Permission is hereby granted, free of charge, to any person obtaining
;; a copy of this software and associated documentation files (the
;; "Software"), to deal in the Software without restriction, including
;; without limitation the rights to use, copy, modify, merge, publish,
;; distribute, sublicense, and/or sell copies of the Software, and to
;; permit persons to whom the Software is furnished to do so, subject to
;; the following conditions:

;; The above copyright notice and this permission notice shall be
;; included in all copies or substantial portions of the Software.

;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :rclmathapi)

;; (cffi:defcfun "Rf_d1mach" :double (a :int))
;; (cffi:defcfun "Rf_i1mach" :double (a :int))
;; #-linux (cffi:defcfun "Rf_gamma_cody" :double (a :double))

;; (cffi:defcfun ("log" rlog) :double (x :double))

;; http://cran.r-project.org/doc/manuals/R-exts.html#Numerical-analysis-subroutines

;; 6.7.3 Numerical Utilities

;; There are a few other numerical utility functions available as
;; entry points.

(cffi:defcfun ("R_pow" pow) :double (x :double) (y :double))
(cffi:defcfun ("R_pow_di" pow-di)  :double (x :double) (i :int))

;; R_pow(x, y) and R_pow_di(x, i) compute x^y and x^i, respectively
;; using R_FINITE checks and returning the proper result (the same as
;; R) for the cases where x, y or i are 0 or missing or infinite or
;; NaN.

(cffi:defcfun "pythag" :double (a :double) (b :double))

;; pythag(a, b) computes sqrt(a^2 + b^2) without overflow or
;; destructive underflow: for example it still works when both a and b
;; are between 1e200 and 1e300 (in IEEE double precision).

;; remapped in some cases to Rlog1p? 
;; (cffi:defcfun "log1p" :double (x :double)) ;; = log(1+x) {care for small x} 

;; Computes log(1 + x) (log 1 plus x), accurately even for small x,
;; i.e., |x| << 1. 

(cffi:defcfun "log1pmx" :double (x :double)) ;; = log(1+x) - x, {care for small x} 

;; Computes log(1 + x) - x (log 1 plus x minus x), accurately even for
;; small x, i.e., |x| << 1.

;; exists already in some cases?
;; (cffi:defcfun "expm1" :double (x :double)) ;; = exp(x)-1 {care for small x}

;; Computes exp(x) - 1 (exp x minus 1), accurately even for small x,
;; i.e., |x| << 1. 

(cffi:defcfun "lgamma1p" :double (x :double)) ;; = log(gamma(x+1)), small x (0 < x < 0.5)

;; Computes log(gamma(x + 1)) (log(gamma(1 plus x))), accurately even
;; for small x, i.e., 0 < x < 0.5.

(cffi:defcfun "logspace_add" :double (logx :double) (logy :double))
(cffi:defcfun "logspace_sub" :double (logx :double) (logy :double))

;; Compute the log of a sum or difference from logs of terms, i.e., "x
;; + y" as log (exp(logx) + exp(logy)) and "x - y" as log (exp(logx) -
;; exp(logy)), without causing overflows or throwing away too much
;; accuracy.

(cffi:defcfun "imax2" :int (x :int) (y :int))
(cffi:defcfun "imin2" :int (x :int) (y :int))
(cffi:defcfun "fmax2" :double (x :double) (y :double))
(cffi:defcfun "fmin2" :double (x :double) (y :double))

;;  Return the larger (max) or smaller (min) of two integer or double
;;  numbers, respectively.

(cffi:defcfun "sign" :double (x :double))

;;  Compute the signum function, where sign(x) is 1, 0, or -1, when x
;;  is positive, 0, or negative, respectively.

(cffi:defcfun "fsign" :double (x :double) (y :double))

;;  Performs "transfer of sign" and is defined as |x| * sign(y).

(cffi:defcfun "fprec" :double (x :double) (digits :double))

;;  Returns the value of x rounded to digits decimal digits (after the
;;  decimal point).

(cffi:defcfun "fround" :double (x :double) (digits :double))

;;  Returns the value of x rounded to digits significant decimal
;;  digits.

(cffi:defcfun "ftrunc" :double (x :double))

;;  Returns the value of x truncated (to an integer value) towards
;;  zero.

